from calendar import c
import numpy as np
import random
from random import randint

#local import
from .functions import outer_product

class Individual():


    def __init__(self,nom='1',ploidy=2,
                 freq={'A':(0.4,0.6),'B':(0.6,0.4)},
                 d=0,R=0.5,mu=(0.1,0.1),
                 sex_system='XY',gen=0,
                 parents=0):
        '''
        Creates an individual based on population characteristics.
        
        Parameters:
            size (int): population size
            ploidy (int): number of homologous chromosomes. Defaults to 2.
            d (float): Linkage desequilibrium. Defaults to 0.
            R (float): recombination frequency. Defaults to 0.5.
            mu (tuple): mutation rate. Defaults to (0.1,0.1).
            sex_system (str): Sex system. Defaults to 'XY'.
            gen (int): current generation. Defaults to 0.
            parents (tuple,int): individual parents. Defaults to 0 (no parents).
            
        '''
        
        #population attributes
        self.spPloidy = ploidy
        self.d = d
        self.R = R
        self.mu = mu
        self.alFreq = freq
        
        #individual attributes
        self.ide = 'g'+str(gen)+".ID-"+str(nom)
        self.age = 0
        self.parents = parents
        self.sex_system = sex_system
        
        self.__createIndividual()
        

        
    def __createIndividual(self):
        '''
        Initialize variables that are not passed to the initializer
        '''
        self.chromosome = dict()
        self.isMutated = 0
        # si tiene padres, es decir, no esta iniciada de 0...
        if self.parents:
            # self.mating()
            self.mating()
            self.isMutated = self.mutation()
        else:
            self.sex_chromosome = self.generate_sexual_genotype()
            self.chromosome = self.chooseGametes()
            self.gameticFreq()
            
        self.sex = self.getSex()
        self.genotype = self.getGenotype()

            
    def generate_sexual_genotype(self):
        '''Generates a sexual genotype, composed of two corresponding values
        to each of the chromosomes, stores them in the variable sex_chromosome
        
        returns:
            str: sex chromosome for the individual'''
        # sex_chromosome= 'XX'or'XY' / 'ZW' or 'ZZ' / 'XO' or 'XX'
        if randint(0,1)==0:
            return self.sex_system[0]+self.sex_system[0]
        else:
            return self.sex_system
        
    def getSex(self):
        '''Returns sex for given sexual chromosome. 
        
        Returns:
            str: sex (Male/Female)
            '''
            
        if self.sex_chromosome[0] == self.sex_chromosome[1]:
            sexo = 'Female' if self.sex_system != 'ZW' else 'Male'
        else:
            sexo = 'Male'  if self.sex_system != 'ZW' else 'Female'
        return sexo
        
    def getGenotype(self):
        '''Obtain the genotype of the individual
        Returns:
            dict(str:str): genotype of the individual'''
        
        c = [list(self.chromosome[x]) for x in self.chromosome]
        array = np.array(c)
        genotype = [''.join(array[:,i]) for i in range(array.shape[1])]
        
        gen_sorted = [''.join(sorted(g)) for g in genotype]
        
        return gen_sorted

    # Calcula la frecuencia gametica a partir de las frecuencias alelicas y D
    # esto quiza deberia estar en population !!!revisar
    def gameticFreq(self):

        f = self.alFreq
        d = self.d
        fGametes = dict()
        
        #producto externo (outer product)
        freqValues = list(f.values())
        inp = freqValues[0]
        finalDict = dict()
        if len(freqValues)>1:
            fGametes = outer_product(f)
        else:
            word = str(list(f.keys())[0])
            keys = [word,word.lower()]
            
            values = list(freqValues[0])  
                 
            fGametes = dict(zip(keys,values))

        return fGametes

    # Elige un gameto segun su probabilidad (frecuencia)
    def chooseGametes(self):
        """Choose a gamete from gameticFreq keys
        by its given probability in gameticFreq values

        Returns:
           dict(str:str): two chromosomes and its genotype
        """
        chrom = dict()
        fGametes = self.gameticFreq()
        
        gameto =list(fGametes.keys())
        pesos = list(fGametes.values())
        
        for i in range(1,self.spPloidy+1):
            
            chrom['c'+str(i)]= ''.join(random.choices(gameto,
                                            weights=pesos,k=1))

        # self.chromosome = chrom
        return chrom

    
    # metodo dunder
    def __str__(self):
        ind_info = {'Individual id:': self.ide,
                    'sex': self.sex, 
                    'genotype': ' '.join(self.genotype)}
        if self.parents != 0:
            ind_info['suffered mutation']= False if not self.isMutated else self.isMutated
            
        return ('\n'.join(f'{key}: {value}' for key,value in ind_info.items()))
    
    def printParents(self,):
        '''
        Print individual parents
        '''
        
        print(f'''su padre es {self.parents[0].ide}, 
        con genotipo {self.parents[0].chromosome}\n su madre es {self.parents[1].ide}
        con genotipo {self.parents[1].chromosome}''')

    # calculates
    def mating(self):
        ''' Calculates the mating of the individual: parent recombination,
        sexual chromosomes and gametes.'''
        # print(self.chromosome,len(self.chromosome))
        if self.spPloidy > 1:
            r = self.R
            for x in range(len(self.parents)):
                # autosomas
                c1P = self.parents[x].chromosome['c1']
                c2P = self.parents[x].chromosome['c2']
                # metodo de recombinacion
                c1,c2 = Individual.recombination(c1P,c2P,r)              
                self.chromosome['c'+str(x+1)] = random.choice([c1,c2])   
            # cromosomas sexuales sex_chromosome: str XX/XY or ZW/ZZ or XX/X0
            sP = self.parents[0].sex_chromosome
            sM = self.parents[1].sex_chromosome
            self.sex_chromosome = ''.join(sorted([random.choice(sP),random.choice(sM)]))
        else:  
            self.chromosome = {'c'+str(k+1):v for k in range(2) for v in random.choice(self.parents[k].chromosome.values())}

    def recombination(c1,c2,r):
        '''Recombination of two chromosomes by r value.
        If r=0, the chromosomes are not recombined.'''
            # factor de interferencia
            # para que sea menos probable la recombinacion de dos loci seguidos
        int = 0
        c1 = list(c1)
        c2 = list(c2)
        c1R,c2R = c1.copy(),c2.copy()
        for i in range(0,len(c1)):
            if random.random()<=r-int:
                # ocurre la recombinacion (solo recombinan los loci A B )
                c1R = c1
                c2R = c2
                c1R[i] = c2[i]
                c2R[i] = c1[i]
                int = r/2
                
            else:
                int = 0      
        return ''.join(c1R),''.join(c2R)
            
    
    def mutation(self):
        '''
        It causes the change from the major to the minor allele with a mut frequency.
        '''
        muType = 'unidirectional'
        muLoci = list()
        if sum(self.mu) != 0:      
            for k,v in self.chromosome.items():
                for i in range(len(v)):
                    #comprueba si es el alelo mayor
                    if v[i].isupper():
                        #si muta, cambia el alelo del cromosoma por el alelo menor
                        if random.random() < self.mu[i]:
                            self.chromosome[k] = self.chromosome[k].replace(v[i],v[i].lower())
                            muLoci.append(v[i])
        
        return muLoci
                    

            





import random
from struct import pack

def outer_product(f: dict):
    """Calculate the outer product of two vectors and stores them on a dictionary
        Args:
        f(dict[char:str]): chromosome
        
        Returns:
        dict: genotype"""
    # lista de tuplas con los valores de frecuencia alelica
    fValues = list(f.values())
    # primera tupla
    a = fValues[0]
    finalD = dict()
    # recorremos todos los valores
    for x in range(1,len(fValues)):
        d = dict()
        values = list()
        names = list()
        k=0
        # recorremos la tupla numero i
        for i,val_i in enumerate(a):
            #recorremos el resto de tuplas
            for j,val_j in enumerate(fValues[x]):
                    # multiplicamos el valor i de la primera tupla
                    # por el valor j de la tupla x
                    values.append(val_i*val_j)
                    # cambiamos el nombre
                    letrai,letraj = rename(i,j,x,k,finalD)
                    names.append(letrai+letraj)
                    if j%2==1:
                        k+=1
        finalD = dict(zip(names,values))
        a = list(finalD.values())
    if len(list(f.keys()))==1: # si solo hay un locus
        # la letra (prob. A)
        letter = list(f.keys())
        # lista de keys (A,a)
        letter.append(letter[0].lower())
        # lista de valores ej: (0.4,0,6)
        values = list(f.values())[0]
        # se hace diccionario
        finalD = dict(zip(letter,values))
        
    return finalD

def rename(i,j,iteration,k,finalDict):
    """rename

    Args:
        i (_type_): _description_
        j (_type_): _description_
        iteration (_type_): _description_
        k (_type_): _description_
        finalDict (_type_): _description_

    Returns:
        _type_: _description_
    """
    # check if it is the first round
    if iteration==1:
        letrai =chr(ord('A')+iteration-1)
        # check if its the first position
        if i==1:
            letrai = letrai.lower()
    else:
        letrai = list(finalDict.keys())[k]

    if j==1:
        letraj = chr(ord('A')+iteration).lower()
    else:
        letraj = chr(ord('A')+iteration)
    return letrai,letraj


def fitness(fit : dict, genotype):
    '''
    Given selection type (fit) and the genotype that is passed
    Returns you whether that individual lives or dies.
    
    Args:
        fit ( dict[str : int] ): genotype or allele fitness value
        ex.1: fit = {'locusA':(0.8,0.9)...}
        ex.2. fit = {'}
        
        child_gen (dict[str:int]): genotype of the individual
    '''
    def is_alive(p):
        """
        Life or death function, if the value generated
        is greater than the value passed as the die parameter.

        Args:
            p(int): probability of death

        returns:
            bool: decide whether to live(True) or die(False)
        """
        randNum = random.random()
        if randNum <= p:
            return True
        else:
            return False

    # Fitness segun locus/genotipo
    # si tiene 1 A = fit[A][0]*fit[A][1]
    # si tiene 0 A = fit[A][1]**2
    # ...
    
    if type(fit) is not int:
        genotipo = ''.join(genotype)
        for k,v in fit.items():
            if len(k)>1:
                if k in genotipo:
                    return is_alive(v)
                else:
                    return True 
            else:
                occ = genotipo.count(k)
                if occ != 0:
                    if isinstance(v, tuple):
                        viveTuple = [is_alive(i**occ) for i in v]
                        vive = all(viveTuple)
                    else:
                        vive = is_alive(v**occ)
                        
                    return vive
                else:
                    return True
            
    # fitness numerico
    if fit==0:
        return True
    if fit==1:
        if 'Aa'== genotype['A']:
            return is_alive(0.9)
        if 'aa'== genotype['A']:
            return is_alive(0.81)
    if fit==2:
        if 'Aa'== genotype['A']:
            return is_alive(0.9)
        if 'AA'== genotype['A']:
            return is_alive(0.81)
    if fit==3:
        if 'aa'== genotype['A'] and 'bb'== genotype['B']:
            return False
    
    return True
        
        
    




import pandas as pd 
import matplotlib.pyplot as plt 
import matplotlib
import os
import numpy as np

from .plot import Plot

#local import
from .population import Population

class Superpop():
    
    def __init__(self,popsize,n,**kwargs):
        '''
        Multiple population simulations at the same time.
        
        Args:
            popsize (int): population size for all populations
            n (int): number of populations
            kwargs: same as Population class
                ploidy (int): Number of homologous chromosomes. Defaults to 2.
                R (float) : Recombination frequency [0,0.5] where 0.5 indicates
                statistic independence. Defaults to 0.5. 
                mu (tuple(float,float)): Mutation rate. Defaults to (1e-4,1e-4).
                freq (dict): loci (key) and allelic frequencies for each allele (values).
                D (float): initial linkage desequilibrium [0,0.5]. Defaults to 0. 
                fit (int,dict): fitness function applied, can take allele fitness value
                or genotype fitness value. E.g. {'A':0.8} or {'AABB':0.8}. Every other 
                value is set to 1. Defaults to 0 (no fitness function applied).
                sex_system (str) : sex determination sistem. Can be XY,ZW or X0. 
                Defaults to 'XY'.
                rnd (bool) : If set to True every other parameter is changed to
                a new random value (R,D,mu,freq). Defaults to False.
            '''

        self.popsize = popsize
        self.n = n
        self.kwargs = kwargs
        self.sPop = [Population(popsize,**kwargs)for x in range(n)] 
        # inicializamos
        [x.initIndividuals()  for x in self.sPop]
    
        
    def evolvePops(self,gens,**kwargs):
        """Inicializes population individuals and evolves them

        Args:
            gens (int): number of generations

        Returns:
            list: list of dictionaries which contains allelic frequencies 
            for every population
        """
        cada = 5
        for key,value in kwargs.items():
            if key == 'every':
                cada = value

            
        printInfo = False
        [x.evolvePop(gens,every=cada,printInfo=False) for x in self.sPop]
        # guarda las frecuencias alelicas acumuladas en una lista
        self.freqs = [d.f_ale_acc for d in self.sPop]
        os.system("cls")
        
        return self.freqs
    
    def plotPops(self):
        """Displays allelic frequencies with matplotlib
        """
    
        # numero de generaciones
        gens = self.sPop[0].gen
        # recogida de informacion
        steps = self.sPop[0].steps
        #ploidia
        ploidy = self.sPop[0].ploidy
        
        labels = [x for x in range(0,gens+1,steps)]
        
              
        popListnames = ['p'+str(i) for i in range(len(self.freqs))]
        
        fig,ax=plt.subplots(1,ploidy,figsize=(13,6))
        plt.suptitle(f"Genetic drift for {self.popsize} individuals")
        plt.style.use('ggplot')
        if "rnd" in self.kwargs and self.kwargs["rnd"] == True:
            caption = ""
        else:
            caption=f"""Caracteristicas iniciales: frecuencia gametica={self.sPop[0].freq}, R={self.sPop[0].R}
                    frecuencia de mutación={self.sPop[0].mu}""" 
        # plt.figtext(0.5, 0.01, caption, wrap=True, horizontalalignment='center', fontsize=10)
        for j,let in enumerate(self.freqs[0].keys()):
            for i in range(len(self.freqs)): 
                with plt.style.context("ggplot"):
                    ax[j].set_title(f"Locus {let}")
                    ax[j].set_ylabel(f'p({let})')
                    ax[j].set_xlabel("Generations")
                    ax[j].plot(self.freqs[i][let])
                    ax[j].set_ylim([0,1])
        new_steps = int(len(labels)/5) if len(labels)>8 else 1
        plt.setp(ax, xticks=range(0,len(labels),new_steps), xticklabels=labels[::new_steps])
        
        plt.show()
                    
                    
        
        
        
if __name__ == '__main__':
    
    superpop = Superpop(10,n=5,ploidy=2,fit={'A':(0.5,0.5)},rnd=False)
    
    alFreq_sp = superpop.startPops(gens=100,every=1)
    
    
    superpop.plotPops()
        
